
# THIS FILE IS GENERATED FROM OPENMM SETUP.PY
short_version = '7.5.0'
version = '7.5.0'
full_version = '7.5.0.dev-b49b82e'
git_revision = 'b49b82efb5a253a7c891ca084b3370e181de2ea3'
release = False
openmm_library_path = r'D:/bld/openmm_1607620238486/_h_env/Library/lib'

if not release:
    version = full_version
